from .detailed_suppl_table import DetailedSupplTable
from .focus_count_table import FocusCountTable
from .phenopacket_ingestor import PhenopacketIngestor
from .simple_patient import SimplePatient
from .simple_variant import SimpleVariant

from .hpo_category import HpoCategorySet


from collections import defaultdict
from typing import List
import pandas as pd
from .hpo_cr import HpoConceptRecognizer


class ColumnMapper:
    def __init__(self) -> None:
        pass


    def map_cell(self, cell_contents) -> List:
        raise NotImplementedError("Need to implement a subclass of ColumnMapper for map_cell")     


    def preview_column(self, column):
        raise NotImplementedError("Need to implement a subclass of ColumnMapper for preview_column")    
   
            
        


from .age_column_mapper import AgeColumnMapper
from .case_encoder import CaseEncoder
from .cohort_encoder import CohortEncoder
from .column_mapper import ColumnMapper
from .custom_column_mapper import CustomColumnMapper
from .hpo_cr import HpoConceptRecognizer
from .hpo_parser import HpoParser
from .hp_term import HpTerm
from .individual import Individual
from .metadata import MetaData
from .option_column_mapper import OptionColumnMapper
from .sex_column_mapper import SexColumnMapper
from .simple_column_mapper import SimpleColumnMapper
from .thresholded_column_mapper import ThresholdedColumnMapper
from .variant_validator import VariantValidator
from .variant_column_mapper import VariantColumnMapper
from .variant import Variant



import unittest
import os
import re
from pyphetools.creation import AgeColumnMapper

HP_JSON_FILENAME = os.path.join(os.path.dirname(__file__), 'data', 'hp.json')
# The API requires us to pass a column name but the column name will not be used in the tests
TEST_COLUMN = "test"

class TestOptionMapper(unittest.TestCase):

    def test_year(self):
        ageMapper = AgeColumnMapper.by_year(column_name=TEST_COLUMN)
        p3 = ageMapper.map_cell("3")
        self.assertEqual("P3Y", p3)
        p42 = ageMapper.map_cell("42")
        self.assertEqual("P42Y", p42)
        
    def test_iso8601(self):
        ageMapper = AgeColumnMapper.iso8601(column_name=TEST_COLUMN)
        p3y = ageMapper.map_cell("P3Y")
        self.assertEqual("P3Y", p3y)
        p3m25d = ageMapper.map_cell("P3Y2M5D")
        self.assertEqual("P3Y2M5D", p3m25d)
        
    def test_year_month_both_1(self):
        ageMapper = AgeColumnMapper.by_year_and_month(column_name=TEST_COLUMN)
        age_string = "14 y 8 m"
        age_iso = ageMapper.map_cell(age_string)
        self.assertEqual("P14Y8M", age_iso)
        
    def test_year_month_both_2(self):
        ageMapper = AgeColumnMapper.by_year_and_month(column_name=TEST_COLUMN)
        age_string = "7 y 6 m"
        age_iso = ageMapper.map_cell(age_string)
        self.assertEqual("P7Y6M", age_iso)
        
    def test_year_month_both_3(self):
        ageMapper = AgeColumnMapper.by_year_and_month(column_name=TEST_COLUMN)
        age_string = "7y6m"
        age_iso = ageMapper.map_cell(age_string)
        self.assertEqual("P7Y6M", age_iso)
        
    def test_year_month_just_year_1(self):
        ageMapper = AgeColumnMapper.by_year_and_month(column_name=TEST_COLUMN)
        age_string = "7 y"
        age_iso = ageMapper.map_cell(age_string)
        self.assertEqual("P7Y", age_iso)
        
    def test_year_month_just_month_1(self):
        ageMapper = AgeColumnMapper.by_year_and_month(column_name=TEST_COLUMN)
        age_string = "2 m"
        age_iso = ageMapper.map_cell(age_string)
        self.assertEqual("P2M", age_iso)

    def test_int_or_float_regex(self):
        int_or_float = r"(\d+)(\.\d+)?"
        p=re.compile(int_or_float)
        y_only = "42"
        results = p.search(y_only).groups()
        self.assertEqual(2, len(results))
        self.assertEqual('42', results[0])
        self.assertIsNone(results[1])
        y_and_m = "42.25"
        results = p.search(y_and_m).groups()
        self.assertEqual(2, len(results))
        self.assertEqual('42', results[0])
        self.assertEqual('.25', results[1])

    def test_fractional_year_strings(self):
        ageMapper = AgeColumnMapper.by_year(column_name=TEST_COLUMN)
        age_string = "2"
        age_iso = ageMapper.map_cell(age_string)
        self.assertEqual("P2Y", age_iso)
        age_string = "4.75"
        age_iso = ageMapper.map_cell(age_string)
        self.assertEqual("P4Y9M", age_iso)
        age_string = "5.9"
        age_iso = ageMapper.map_cell(age_string)
        self.assertEqual("P5Y11M", age_iso)
        age_string = "6.25"
        age_iso = ageMapper.map_cell(age_string)
        self.assertEqual("P6Y3M", age_iso)
        age_string = "8.1"
        age_iso = ageMapper.map_cell(age_string)
        self.assertEqual("P8Y1M", age_iso)




import os
import random
import time
import phenopackets
from phenopackets import Phenopacket
from google.protobuf.json_format import Parse
from google.protobuf.json_format import MessageToJson
from google.protobuf.message import *
import json


class Downsampler:
    """
    This class flattens all observed terms into a set and also recorded variants, sex, identifier, and age
    The class purposefully disregards information about the time course in order to be able to count the 
    frequencies of HPO terms in groups
    """

    def __init__(self, phenopacket_file=None, ga4gh_phenopacket=None) -> None:
        if phenopacket_file is None and ga4gh_phenopacket is None:
            raise ValueError("Must pass either 'phenopacket_file' or 'ga4gh_phenopacket' argument")
        elif phenopacket_file is not None and ga4gh_phenopacket is not None:
            raise ValueError("Must pass only one of 'phenopacket_file' and 'ga4gh_phenopacket' arguments")
        elif phenopacket_file is not None:
            if not os.path.isfile(phenopacket_file):
                raise FileNotFoundError(f"Could not find phenopacket file at '{phenopacket_file}'") 
            with open(phenopacket_file) as f:
                data = f.read()
                jsondata = json.loads(data)
                self._phenopacket = Parse(json.dumps(jsondata), Phenopacket())
        else:
            # in this case, ga4gh_phenopacket cannot be None
            if str(type(ga4gh_phenopacket)) != "<class 'phenopackets.schema.v2.phenopackets_pb2.Phenopacket'>":                   
                raise ValueError(f"phenopacket argument must be GA4GH Phenopacket Schema Phenopacket but was {type(ga4gh_phenopacket)}")
            else:
                self._phenopacket = ga4gh_phenopacket

    def downsample_positive_hpo_terms(self, k) -> Phenopacket:
        if not isinstance(k, int):
            raise ValueError(f"argument k must be an integer but was {type(k)}")
        observed_hpo_terms = []
        for pf in self._phenopacket.phenotypic_features:
            if pf.excluded:
                continue
            else:
                observed_hpo_terms.append(pf)
        downsampled = random.sample(observed_hpo_terms, k)
        print(f"We downsampled {len(observed_hpo_terms)} to {len(downsampled)} terms")
        downsampled_ppacket = Phenopacket()
        downsampled_ppacket.MergeFrom(self._phenopacket)
        downsampled_ppacket.phenotypic_features.clear()
        downsampled_ppacket.phenotypic_features.MergeFrom(downsampled)
        return downsampled_ppacket

    def write_downsampled_phenopacket(self, k, basename, outdir=None) -> None:
        downsampled_ppacket = self.downsample_positive_hpo_terms(k=k)
        json_string = MessageToJson(downsampled_ppacket)
        fname = basename + f"_{k}_{time.time()}_.json"
        if outdir is not None:
            outpth =  os.path.join(outdir, basename, fname)
        else:
            outpth = os.path.join(basename, fname)
        with open(outpth, "wt") as fh:
            fh.write(json_string)
            print(f"Wrote phenopacket to {outpth}")

        
        
    





import json
from collections import defaultdict
from .hpo_cr import HpoConceptRecognizer
from .hpo_exact_cr import HpoExactConceptRecognizer
import urllib.request 
import os


HPO_JSON_URL = "https://raw.githubusercontent.com/obophenotype/human-phenotype-ontology/master/hp.json"


def extract_curie(url):
    if not url.startswith('http://purl.obolibrary.org/obo/HP'):
        return None
    return url[31:].replace("_", ":")

class Edge:
    """_summary_
    This class represents an is-a link in the HPO graph.
    Note that instances of this class are not intended for use outside of parsing, where we use them to 
    restrict the terms that are return to descendants of Phenotypic abnormality (HP:0000118).
    """
    def __init__(self, e) -> None:
        self._subject = extract_curie(e.get('sub'))
        self._object = extract_curie(e.get('obj'))
        
    def is_valid(self):
        return self._subject is not None and self._object is not None
    
    @property
    def subject(self):
        return self._subject
    
    @property
    def object(self):
        return self._object

def is_phenotype(hpo_id, subject_to_edge_d):
    """_summary_
    return True if hpo_id is a descendant of HP:0000118, Phenotypic abnormality, otherwise return False
    Args:
        hpo_id (str): a candidate HPO term
        subject_to_edge_d (dict): subject: HPO CURIE; object: list of Edges with this CURIE as subject

    Returns:
        bool: return True if hpo_id is a descendant of HP:0000118, otherwise False
    """
    stack = [hpo_id]
    while len(stack) > 0:
        id = stack.pop()
        parent_edges = subject_to_edge_d.get(id, [])
        for p in parent_edges:
            if p.object == "HP:0000118":
                return True
            stack.append(p.object)
    return False
   
    
def extract_phenotype_descendant_ids(nodes, edges):
    """_summary_
    Create a simple graph structure and extract those HPO terms that are descendants
    of the Phenotypic abnormality
    Returns:
        Set: set of valid HPO terms describing only phenotypic abnormalities
    """
    # subject: HPO CURIE; object: list of Edges with this CURIE as subject
    subject_to_edge_d = defaultdict(list)
    for e in edges:
        edge = Edge(e)
        if edge.is_valid():
            subject_to_edge_d[edge.subject].append(edge)
    valid_nodes = set() 
    for n in nodes:
        node_id = extract_curie(n.get('id'))
        if node_id is not None and node_id.startswith("HP:"):
            if is_phenotype(node_id, subject_to_edge_d): 
                valid_nodes.add(node_id)
    return valid_nodes


class HpoParser:
    """
    Extract maps of labels and ids from the hp.json file.
    Note that we do not need to have the ontological structure of the HPO for this
    application and so only nodes and version are extracted.
    """
    def __init__(self, hpo_json_file=None):
        if hpo_json_file is None:
            hpo_json_file = self._download_to_disk_if_necessary()
        with open(hpo_json_file) as f:
            data = json.load(f)  
            graphs = data.get('graphs')
        if len(graphs) != 1:
            raise ValueError(f"Expected to get one graph but got {len(graphs)}")
        hpo = graphs[0]
        id = hpo.get("id")
        if id != "http://purl.obolibrary.org/obo/hp.json":
            raise ValueError(f"Expected to get hp.json but got {id}")  
        nodes = hpo.get("nodes") 
        edges = hpo.get("edges")
        valid_node_curies = extract_phenotype_descendant_ids(nodes=nodes, edges=edges)  
        id_to_primary_label = defaultdict()
        label_to_id = defaultdict()
        for n in nodes:
            if 'id' in n and 'HP_' in n.get('id'):
                hpo_id, label, all_labels = self._extract_node_data(n)
                if not hpo_id in valid_node_curies:
                    continue # This restricts us to descendants of Phenotypic abnormality
                id_to_primary_label[hpo_id] =  label
                for lab in all_labels:
                    label_to_id[lab.lower()] = hpo_id
        self._id_to_primary_label = id_to_primary_label
        self._label_to_id = label_to_id
        meta = hpo.get('meta')
        long_version = meta.get('version')
        if 'http://purl.obolibrary.org/obo/hp/releases/' in long_version:
            version = long_version[43:].split('/')[0]
        else:
            version = long_version
        self._long_version = long_version
        self._version = version

    def _extract_node_data(self, json_node):
        if not isinstance(json_node, dict):
            raise ValueError("HpoNode obhject must be constructed from JSON-derived dictionary")
        if not 'id' in json_node.keys():
            raise ValueError("HpoNode object must have id attribute")
        id = json_node.get('id')
        if not 'lbl' in json_node.keys():
            raise ValueError(f"HpoNode object must have lbl attribute but {id} did not")
        if not 'http://purl.obolibrary.org/obo/HP' in id:
            raise ValueError(f"HpoNode id must begin with 'http://purl.obolibrary.org/obo/' but we got {id}")
        hpo_id = id[31:].replace("_", ":")
        label = json_node.get("lbl","NA")
        all_labels = set()
        all_labels.add(label)
        # synonyms are nest
        metadata = json_node.get('meta')
        if metadata is not None:
            synonyms = metadata.get('synonyms', None)
            if synonyms is not None:
                for syn in synonyms:
                    synonym = syn.get('val')
                    if len(synonym) < 5:
                        continue
                    all_labels.add(synonym)
        return hpo_id, label, list(all_labels)
    
    def _download_to_disk_if_necessary(self):
        local_dir = "hpo_data"
        if not os.path.exists(local_dir):
            os.makedirs(local_dir)
        hpo_json_file = os.path.join(local_dir, "hp.json")
        if not os.path.isfile(hpo_json_file):
            urllib.request.urlretrieve(HPO_JSON_URL, hpo_json_file)
        return hpo_json_file
        

    def get_version(self):
        return self._version
    
    def get_hpo_concept_recognizer(self) -> HpoConceptRecognizer:
        return HpoExactConceptRecognizer(label_to_id=self._label_to_id, id_to_primary_label=self._id_to_primary_label)


from hpotk.model import TermId


class TermCount:

    def __init__(self, id) -> None:
        if not isinstance(id, TermId):
            raise ValueError(f"id argument must be TermId but was {type(id)}")
        self._hpo_id = id
        self._count = 0

    def increment(self):
        self._count += 1

    @property
    def hpo_id(self):
        return self._hpo_id

    @property
    def count(self):
        return self._count


import os
import phenopackets
from phenopackets import Phenopacket
from google.protobuf.json_format import Parse
import json
from collections import defaultdict

from ..creation.hp_term import HpTerm
from .simple_variant import SimpleVariant

class SimplePatient:
    """
    This class flattens all observed terms into a set and also recorded variants, sex, identifier, and age
    The class purposefully disregards information about the time course in order to be able to count the 
    frequencies of HPO terms in groups
    """

    def __init__(self, phenopacket_file=None, ga4gh_phenopacket=None) -> None:
        if phenopacket_file is None and ga4gh_phenopacket is None:
            raise ValueError("Must pass either 'phenopacket_file' or 'ga4gh_phenopacket' argument")
        elif phenopacket_file is not None and ga4gh_phenopacket is not None:
            raise ValueError("Must pass only one of 'phenopacket_file' and 'ga4gh_phenopacket' arguments")
        elif phenopacket_file is not None:
            if not os.path.isfile(phenopacket_file):
                raise FileNotFoundError(f"Could not find phenopacket file at '{phenopacket_file}'") 
            with open(phenopacket_file) as f:
                data = f.read()
                jsondata = json.loads(data)
                ppack = Parse(json.dumps(jsondata), Phenopacket())
        else:
            # in this case, ga4gh_phenopacket cannot be None
            if str(type(ga4gh_phenopacket)) != "<class 'phenopackets.schema.v2.phenopackets_pb2.Phenopacket'>":                   
                raise ValueError(f"phenopacket argument must be GA4GH Phenopacket Schema Phenopacket but was {type(phenopacket)}")
            else:
                ppack = ga4gh_phenopacket
        observed_hpo_terms = defaultdict(HpTerm)
        excluded_hpo_terms = defaultdict(HpTerm)
        self._phenopacket_id = ppack.id
        if ppack.subject is None:
            raise ValueError("Phenopackets must have a subject message to be used with this package")
        subj = ppack.subject
        if subj.id is None:
            raise ValueError("Phenopacket subjects must have an id field to be used with this package")
        else:
            self._subject_id = subj.id
        if subj.time_at_last_encounter is None or subj.time_at_last_encounter.age.iso8601duration is None:
            print("Warning: No age found for phenopacket")
            self._time_at_last_encounter = None
        else:
            self._time_at_last_encounter = subj.time_at_last_encounter.age.iso8601duration 
        if ppack.subject.sex == phenopackets.MALE:
            self._sex = "MALE"
        elif ppack.subject.sex == phenopackets.FEMALE:
            self._sex = "FEMALE"
        elif ppack.subject.sex == phenopackets.OTHER_SEX:
            self._sex = "OTHER"
        else:
            self._sex = "UNNOWN"
        for pf in ppack.phenotypic_features:
            hpterm =  HpTerm(id=pf.type.id, label=pf.type.label, observed=not pf.excluded)
            if pf.excluded:
                excluded_hpo_terms[pf.type.id] = hpterm
            else:
                observed_hpo_terms[pf.type.id] = hpterm
        for k, v in observed_hpo_terms.items():
            if k in excluded_hpo_terms:
                excluded_hpo_terms.pop(k) # remove observed terms that may have been excluded at other occasion
        self._observed = observed_hpo_terms
        self._excluded = excluded_hpo_terms
        # Add information about variants
        self._variant_d = defaultdict(SimpleVariant)
        if ppack.interpretations is not None and len(ppack.interpretations) > 0:
            interprets = ppack.interpretations
            for intp in interprets:
                variant = SimpleVariant(interpretation=intp)
                self._variant_d[variant.hgvs] = variant
        # TODO -- add information about disease, ensuring that all Variants are associated with the same diagnosis
                
        


    def get_phenopacket_id(self):
        return self._phenopacket_id

    def get_subject_id(self):
        return self._subject_id

    def get_sex(self):
        return self._sex

    def get_age(self):
        return self._time_at_last_encounter

    def get_observed_hpo_d(self):
        """
        returns map with key (string) HP id, value, HpTerm from creation submodule
        """
        return self._observed

    def get_excluded_hpo_d(self):
        return self._excluded

    def get_variant_d(self):
        return self._variant_d



## constants.py
## used to define constants that are used in multiple places in the package
## these constants should not be exported



MALE_SYMBOL = 'MALE'
FEMALE_SYMBOL = 'FEMALE'
OTHER_SEX_SYMBOL = 'OTHER'
UNKOWN_SEX_SYMBOL = 'UNKNOWN'


import unittest
import os
from pyphetools.creation import HpoParser

HP_JSON_FILENAME = os.path.join(os.path.dirname(__file__), 'data', 'hp.json')

class TestHpoParser(unittest.TestCase):
    
    @classmethod
    def setUpClass(cls) -> None:
        parser = HpoParser(hpo_json_file=HP_JSON_FILENAME)
        cls.hpo_cr = parser.get_hpo_concept_recognizer()
        
    def test_not_none(self):
        self.assertIsNotNone(self.hpo_cr)



import pandas as pd
from math import isnan
from typing import List
import os
import phenopackets
from google.protobuf.json_format import MessageToJson
import re

from .column_mapper import ColumnMapper
from .variant_column_mapper import VariantColumnMapper
from .hpo_cr import HpoConceptRecognizer
from .individual import Individual
from .metadata import MetaData

class CohortEncoder:
    
    def __init__(self, df, hpo_cr, column_mapper_d, individual_column_name, agemapper, sexmapper, metadata, variant_mapper=None, pmid=None):
        if not isinstance(df, pd.DataFrame):
            raise ValueError(f"df argument must be pandas data frame but was {type(df)}")
        if not isinstance(hpo_cr, HpoConceptRecognizer):
            raise ValueError(f"hpo_cr argument must be a HpoConceptRecognizer but was {type(hpo_cr)}")
        if not isinstance(column_mapper_d, dict):
            raise ValueError(f"column_mapper_d argument must be a dictionary but was {type(column_mapper_d)}")
        if not isinstance(individual_column_name, str):
            raise ValueError(f"individual_column_name argument must be a string but was {type(individual_column_name)}")
        if variant_mapper is not None and not isinstance(variant_mapper, VariantColumnMapper):
            raise ValueError(f"variant_mapper argument must be VariantColumnMapper but was {type(variant_mapper)}")
        if metadata is None:
            raise ValueError("Must pass a metadata object to constructor")
        # the following workaround is needed because isinstance gets confused about the two following classes
        elif str(type(metadata)) == "<class 'phenopackets.schema.v2.core.meta_data_pb2.MetaData'>":
            self._metadata = metadata
        elif str(type(metadata)) == "<class 'pyphetools.creation.metadata.MetaData'>":
            self._metadata = metadata.to_ga4gh()
        else:
            raise ValueError(F"Malformed metadata argument of type {type(metadata)}")
        self._df = df
        self._hpo_concept_recognizer = hpo_cr
        self._column_mapper_d = column_mapper_d
        self._id_column_name = individual_column_name
        #self._sex_column = individual_d.get('sex')
        self._age_mapper = agemapper
        self._sex_mapper = sexmapper
        self._disease_id = None
        self._disease_label = None
        self._variant_mapper = variant_mapper
        self._pmid = pmid
        
    def preview_dataframe(self):
        """
        Generate a dataframe with a preview of the parsed contents
        """
        df = self._df.reset_index()  # make sure indexes pair with number of rows
        individuals = []
        age_column_name = self._age_mapper.get_column_name()
        sex_column_name = self._sex_mapper.get_column_name()
        for index, row in df.iterrows():
            individual_id = row[self._id_column_name]
            age_cell_contents = row[age_column_name]
            age = self._age_mapper.map_cell(age_cell_contents)
            sex_cell_contents = row[sex_column_name]
            sex = self._sex_mapper.map_cell(sex_cell_contents)
            hpo_terms = []
            for column_name, column_mapper in self._column_mapper_d.items():
                cell_contents = row[column_name]
                ## Empty cells are often represented as float non-a-number by Pandas
                if isinstance(cell_contents, float) and isnan(cell_contents):
                    continue
                terms = column_mapper.map_cell(row[column_name])
                hpo_terms.extend(terms)
            hpo_string = "\n".join([h.to_string() for h in hpo_terms])
            d = {'id': individual_id,
                 'sex': sex,
                 'age': age,
                 'phenotypic features': hpo_string}
            individuals.append(d)
        df = pd.DataFrame(individuals)
        return df.set_index('id')
    
    def set_disease(self, disease_id, label):
        """_summary_
        If all patients in the cohort have the same disease we can set it with this method
        Args:
            id (str): disease identifier, e.g., OMIM:600123
            label (str): disease name
        """
        self._disease_id = disease_id
        self._disease_label = label
    
    
    def get_individuals(self, additional_hpo=None) -> List[Individual]:
        df = self._df.reset_index()  # make sure indexes pair with number of rows
        individuals = []
        age_column_name = self._age_mapper.get_column_name()
        sex_column_name = self._sex_mapper.get_column_name()
        if self._variant_mapper is None:
            variant_colname = None
        else:
            variant_colname = self._variant_mapper.get_column_name()
        count = 0
        for index, row in df.iterrows():
            individual_id = row[self._id_column_name]
            age_cell_contents = row[age_column_name]
            age = self._age_mapper.map_cell(age_cell_contents)
            sex_cell_contents = row[sex_column_name]
            sex = self._sex_mapper.map_cell(sex_cell_contents)
            if additional_hpo is None:
                hpo_terms = []
            else:
                hpo_terms = additional_hpo[count]
            for column_name, column_mapper in self._column_mapper_d.items():
                if column_name not in df.columns:
                    raise ValueError(f"Did not find column name '{column_name}' in dataframe -- check spelling!")
                cell_contents = row[column_name]
                ## Empty cells are often represented as float non-a-number by Pandas
                if isinstance(cell_contents, float) and isnan(cell_contents):
                    continue
                terms = column_mapper.map_cell(cell_contents)
                hpo_terms.extend(terms)
            if variant_colname is not None:
                variant_col = row[variant_colname]
                variant_list = self._variant_mapper.map_cell(variant_col)
            else:
                variant_list = []
            indi = Individual(individual_id=individual_id, sex=sex, age=age, hpo_terms=hpo_terms, variant_list=variant_list,
                              disease_id=self._disease_id, disease_label=self._disease_label)
            individuals.append(indi)
            count += 1
        return individuals

    def output_phenopackets(self, outdir="phenopackets"):
        """Output data about all individuals as GA4GH phenopackets

        Args:
            outdir (str): name of directory to write phenopackets (default: 'phenopackets')
        """
        if not os.path.exists(outdir):
            os.makedirs(outdir)
        individual_list = self.get_individuals()
        written = 0
        for individual in individual_list:
            phenopckt = individual.to_ga4gh_phenopacket(metadata=self._metadata)
            json_string = MessageToJson(phenopckt)
            if self._pmid is None:
                fname = "phenopacket_" + individual.id + ".json"
            else:
                pmid = self._pmid.replace(" ", "").replace(":","_")
                fname = pmid + "_" + individual.id + ".json"
            fname = re.sub('[^\w_.)( -]', '', fname) #remove any illegal characters from filename
            fname = fname.replace(" ", "_")
            outpth = os.path.join(outdir, fname)
            with open(outpth, "wt") as fh:
                fh.write(json_string)
                written += 1
        print(f"Wrote {written} phenopackets to {outdir}")



from .variant import Variant
from typing import List
from .variant_validator import VariantValidator
import pandas as pd

ACCEPTABLE_GENOTYPES = {"heterozygous", "homozygous", "hemizygous"}



class VariantColumnMapper:
    
    def __init__(self, assembly, transcript, column_name, genotype=None) -> None:
        """
        genotype -- the 'default' for all variants in this column
        """
        if genotype is not None and genotype not in ACCEPTABLE_GENOTYPES:
            raise ValueError(f"Did not recognize default genotype {genotype}")
        self._genotype = genotype
        if transcript is None or len(transcript) < 3:
            raise ValueError(f"Invalid transcript: \"{transcript}\"")
        self._transcript = transcript
        self._validator = VariantValidator(genome_build=assembly, transcript=transcript)
        self._column_name = column_name
        
    def map_cell(self, cell_contents, delimiter=None) -> List[Variant]:
        if delimiter is not None:
            items = [x.strip() for x in cell_contents.split(delimiter)]
        else:
            items = [cell_contents.strip()]
        results = []
        for item in items:
            try:
                variant = self._validator.encode_hgvs(item)
                if self._genotype is not None:
                    variant.set_genotype(self._genotype)
                results.append(variant)
            except Exception as exc:
                print(f"Not able to get variant for {item}: {exc}")
        return results
    
    def preview_column(self, column) -> pd.DataFrame:
        if not isinstance(column, pd.Series):
            raise ValueError("column argument must be pandas Series, but was {type(column)}")
        dlist = []
        for _, value in column.items():
            variants  = self.map_cell(str(value))
            if len(variants) == 0:
                dlist.append({"variant": "n/a"})
            else:
                result_strings = []
                for v in  variants:
                    result_strings.append(v.to_string()) 
                dlist.append({"variant": ": ".join(result_strings)})
        return pd.DataFrame(dlist)  
    
    def get_column_name(self):
        return self._column_name



import phenopackets 

ACCEBTABLE_GENOMES = {"GRCh37", "GRCh38", "hg19", "hg38"}



# {'alt': 'C', 'chr': '16', 'pos': '1756403', 'ref': 'CG'}},

class Variant:
    """
    This encapsulates variant data that we retrieve from Variant Validator
    """
    def __init__(self, assembly, vcf_d, symbol=None, hgnc=None, hgvs=None, transcript=None, g_hgvs=None) -> None:
        if not assembly in ACCEBTABLE_GENOMES:
            raise ValueError(f"Malformed assembly: \"{assembly}\"")
        self._assembly = assembly
        if not isinstance(vcf_d, dict):
            raise ValueError(f"vcf_d argument must be dictionary")
        self._chr = vcf_d.get('chr')
        self._position = int(vcf_d.get('pos'))
        self._ref = vcf_d.get('ref')
        self._alt = vcf_d.get('alt')
        self._symbol = symbol
        self._hgnc_id = hgnc
        self._hgvs = hgvs
        self._transcript = transcript
        self._g_hgvs = g_hgvs
        self._genotype = None
        
    @property
    def assembly(self):
        return self._assembly
        
    @property
    def chr(self):
        return self._chr
        
    @property
    def position(self):
        return self._position
        
    @property
    def ref(self):
        return self._ref
        
    @property
    def alt(self):
        return self._alt
    
    @property
    def genotype(self):
        return self._genotype
    
    def set_genotype(self, gt):
        self._genotype = gt
        
        
    def __str__(self):
        return f"{self._chr}:{self._position}{self._ref}>{self._alt}"
    
    def to_string(self):
        return self.__str__()
    
    def to_ga4gh(self, acmg=None):
        """
        Transform this Variant object into a "variantInterpretation" message of the GA4GH Phenopacket schema
        """
        vdescriptor = phenopackets.VariationDescriptor()
        if self._hgnc_id is not None and self._symbol is not None:
            vdescriptor.gene_context.value_id = self._hgnc_id
            vdescriptor.gene_context.symbol = self._symbol
        hgvs_expression = phenopackets.Expression()
        if self._hgvs is not None:
            hgvs_expression.syntax = "hgvs.c"
            hgvs_expression.value = self._hgvs
            vdescriptor.expressions.append(hgvs_expression) 
        if self._g_hgvs is not None: 
            hgvs_expression.syntax = "hgvs.g"
            hgvs_expression.value = self._g_hgvs
            vdescriptor.expressions.append(hgvs_expression) 
        vdescriptor.molecule_context =  phenopackets.MoleculeContext.genomic
        if self._genotype is not None:
            if self._genotype == 'heterozygous':
                vdescriptor.allelic_state.id = "GENO:0000135"
                vdescriptor.allelic_state.label = "heterozygous"
            elif self._genotype == 'homozygous':
                vdescriptor.allelic_state.id = "GENO:0000136"
                vdescriptor.allelic_state.label = "homozygous" 
            elif self._genotype == 'hemizygous':
                vdescriptor.allelic_state.id = "GENO:0000134"
                vdescriptor.allelic_state.label = "hemizygous" 
            else:
                print(f"Did not recognize genotype {self._genotype}")  
        vinterpretation = phenopackets.VariantInterpretation() 
        if acmg is not None:
            if acmg.lower() == 'benign':
                vinterpretation.acmgPathogenicityClassification = phenopackets.AcmgPathogenicityClassification.BENIGN
            elif acmg.lower == 'likely benign' or acmg.lower() == 'likely_benign':
                vinterpretation.acmgPathogenicityClassification = phenopackets.AcmgPathogenicityClassification.LIKELY_BENIGN
            elif acmg.lower == 'uncertain significance' or acmg.lower() == 'uncertain_significance':
                vinterpretation.acmgPathogenicityClassification = phenopackets.AcmgPathogenicityClassification.UNCERTAIN_SIGNIFICANCE
            elif acmg.lower == 'likely pathogenic' or acmg.lower() == 'likely_pathogenic':
                vinterpretation.acmgPathogenicityClassification = phenopackets.AcmgPathogenicityClassification.LIKELY_PATHOGENIC
            elif acmg.lower == 'pathogenic' or acmg.lower() == 'pathogenic':
                vinterpretation.acmgPathogenicityClassification = phenopackets.AcmgPathogenicityClassification.PATHOGENIC
            else:
                print(f"Warning- did not recognize ACMG category {acmg}")
        vcf_record = phenopackets.VcfRecord()
        vcf_record.genome_assembly =  self._assembly
        vcf_record.chrom = self._chr
        vcf_record.pos = self._position
        vcf_record.ref = self._ref
        vcf_record.alt = self._alt
        vdescriptor.vcf_record.CopyFrom(vcf_record)
        vinterpretation.variation_descriptor.CopyFrom(vdescriptor)
        return vinterpretation
        


from .hp_term import HpTerm
from .column_mapper import ColumnMapper
from .hpo_cr import HpoConceptRecognizer
from typing import List
import pandas as pd
from collections import defaultdict

class OptionColumnMapper(ColumnMapper):
    
    def __init__(self, concept_recognizer, option_d):
        if not isinstance(concept_recognizer, HpoConceptRecognizer):
            raise ValueError("concept_recognizer argument must be HpoConceptRecognizer but was {type(concept_recognizer)}")
        self._hpo_cr = concept_recognizer
        # Either have self._option_d be an empty dictionary or it must be a valid dictionary
        if option_d is None:
            self._option_d = {}
        else:
            if not isinstance(option_d, dict):
                raise ValueError(f"option_d argument must be dictionary but was {type(option_d)}")
            self._option_d = option_d
        
    def map_cell(self, cell_contents) -> List[HpTerm]:
        chunk = cell_contents.strip()
        results = self._hpo_cr.parse_cell(cell_contents=chunk, custom_d=self._option_d)
        if results is None:
            return []
        else:
            return results
        
    def preview_column(self, column) -> pd.DataFrame:
        if not isinstance(column, pd.Series):
            raise ValueError("column argument must be pandas Series, but was {type(column)}")
        dlist = []
        for _, value in column.items():
            column_val = []
            results  = self.map_cell(str(value))
            if results is None:
                print(f"Got None results for {str(value)}")
                dlist.append({"terms": "n/a"})  
            elif len(results) > 0:
                for hpterm in results:
                    column_val.append(f"{hpterm.id} ({hpterm.label}/{hpterm.display_value})")
                dlist.append({"terms": "; ".join(column_val)})  
            else:
                dlist.append({"terms": "n/a"})  
        return pd.DataFrame(dlist)     
   


from .hp_term import HpTerm
from .column_mapper import ColumnMapper
from typing import List
import pandas as pd
from collections import defaultdict


def get_separate_hpos_from_df(df, hpo_cr):
    """Loop through all the cells in a dataframe or series and try to parse each cell as HPO term.
    Useful when the seperate HPO terms are in the cells themselves.

      Args:
         df (dataframe): dataframe with phenotypic data
         hpo_cr (HpoConceptRecognizer): instance of HpoConceptRecognizer to match HPO term and get label/id

      Returns:
          additional_hpos: list of lists with the additional HPO terms per individual
      """
    additional_hpos = []

    for i in range(len(df)):
        temp_hpos = []
        for y in range(df.shape[1]):
            hpo_term = hpo_cr.parse_cell(df.iloc[i, y])
            if len(hpo_term) > 0:
                temp_hpos.extend(hpo_term)
        additional_hpos.append(list(set(temp_hpos)))
    return additional_hpos


def try_mapping_columns(df, observed, excluded, hpo_cr, preview=True):
    """Try to map the columns in a dataframe by matching the name of the column to correct HPO term.
    Wrapper for SimpleColumnMapper below

    Args:
       df (dataframe): dataframe with phenotypic data
       observed (str): symbol used in table if the phenotypic feature was observed
       excluded (str): symbol used if the feature was excluded
       hpo_cr (HpoConceptRecognizer): instance of HpoConceptRecognizer to match HPO term and get label/id
       preview (bool): whether to print the successfully mapped columns

    Returns:
        simple_mappers (dict): dict with successfully mapped columns
        col_not_found (list): columns that were not mapped
    """
    col_not_found = []
    simple_mappers = defaultdict(ColumnMapper)
    for col in df.columns:
        hpo_term = hpo_cr.parse_cell(col)
        if len(hpo_term) > 0:
            hpo_term = hpo_term[0]
            simple_mappers[col] = SimpleColumnMapper(hpo_id=hpo_term.id,
                                                                hpo_label=hpo_term.label,
                                                                observed=observed,
                                                                excluded=excluded)
            if preview:
                print(simple_mappers[col].preview_column(df[col]))
        else:
            col_not_found.append(col)
    return simple_mappers, col_not_found


class SimpleColumnMapper(ColumnMapper):

    def __init__(self, hpo_id, hpo_label, observed=None, excluded=None, non_measured=None, constant=False):
        """_summary_

        Args:
            hpo_id (_type_): HPO  id, e.g., HP:0004321
            hpo_label (_type_): Corresponding term label
            observed (_type_): symbol used in table if the phenotypic feature was observed
            excluded (_type_): symbol used if the feature was excluded
            non_measured (_type_, optional): symbol used if the feature was not measured or is N/A. Defaults to None.
            constant (bool, optional): If true, all patients have this feature. Defaults to False.
        """
        self._hpo_id = hpo_id
        self._hpo_label = hpo_label
        if observed is None or excluded is None:
            if not constant:
                raise ValueError(
                    "constant argument must be true if not arguments are provided for observed and excluded")
        if constant:
            self._observed = set()
            self._excluded = set()
        else:
            self._observed = observed
            self._excluded = excluded
        self._not_measured = non_measured
        self._constant = constant

    def map_cell(self, cell_contents) -> List[HpTerm]:
        if self._constant:
            return [HpTerm(id=self._hpo_id, label=self._hpo_label)]
        if not isinstance(cell_contents, str):
            raise ValueError(
                f"Error: cell_contents argument ({cell_contents}) must be string but was {type(cell_contents)} -- coerced to string")
        contents = cell_contents.strip()
        if contents in self._observed:
            return [HpTerm(id=self._hpo_id, label=self._hpo_label)]
        elif contents in self._excluded:
            return [HpTerm(id=self._hpo_id, label=self._hpo_label, observed=False)]
        else:
            return [HpTerm(id=self._hpo_id, label=self._hpo_label, measured=False)]

    def preview_column(self, column) -> pd.DataFrame:
        if not isinstance(column, pd.Series):
            raise ValueError("column argument must be pandas Series, but was {type(column)}")
        dlist = []
        for _, value in column.items():
            value = self.map_cell(str(value))
            hpterm = value[0]
            dlist.append({"term": hpterm.hpo_term_and_id, "status": hpterm.display_value})
        return pd.DataFrame(dlist)




import unittest
from pyphetools.creation import HpTerm

class TestHpTerm(unittest.TestCase):

    def test_excluded(self):
        hpterm = HpTerm(id="HP:0001250", label="test label")
        self.assertEqual("HP:0001250", hpterm.id)
        self.assertEqual("test label", hpterm.label)
        self.assertTrue(hpterm.observed)
        self.assertTrue(hpterm.measured)
        hpterm.excluded()
        self.assertFalse(hpterm.observed)










import unittest
import os
import pandas as pd
from pyphetools.creation import HpoParser, CaseEncoder
HP_JSON_FILENAME = os.path.join(os.path.dirname(__file__), 'data', 'hp.json')
# The API requires us to pass a column name but the column name will not be used in the tests
TEST_COLUMN = "test"

class TestCaseParse(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        hpparser = HpoParser(hpo_json_file=HP_JSON_FILENAME)
        cls._hpo_cr = hpparser.get_hpo_concept_recognizer()
        cls._parser = CaseEncoder(concept_recognizer=cls._hpo_cr, pmid="PMID:1")

    def test_chief_complaint(self):
        """
        Expect to get
        HP:0000365 	Hearing impairment 	True 	True
        HP:0001742 	Nasal congestion 	True 	True
        HP:0025267 	Snoring
        """
        vignette = "A 17-mo-old boy presented with progressive nasal obstruction, snoring and hearing loss symptoms when referred to the hospital."
        results = self._parser.add_vignette(vignette=vignette)
        self.assertEqual(3, len(results))
        self.assertTrue(isinstance(results, pd.DataFrame))
        id = results.loc[(results['id'] == 'HP:0000365')]['id'].values[0]
        self.assertEqual("HP:0000365", id)

    def test_excluded(self):
        """
        Currently, our text mining does not attempt to detect negation, but users can enter a set of term labels that are excluded
        In this case, He had no nasal obstruction should lead to excluded/Nasal congestion
        """
        vignette = "He had no nasal obstruction and he stated that his smell sensation was intact."
        excluded = set()
        excluded.add("Nasal congestion")
        results = self._parser.add_vignette(vignette=vignette, excluded_terms=excluded)
        self.assertEqual(1, len(results))
        id = results.loc[(results['id'] == 'HP:0001742')]['id'].values[0]
        self.assertEqual("HP:0001742", id)
        label = results.loc[(results['id'] == 'HP:0001742')]['label'].values[0]
        self.assertEqual("Nasal congestion", label)
        observed = results.loc[(results['id'] == 'HP:0001742')]['observed'].values[0]
        self.assertFalse(observed)
        measured = results.loc[(results['id'] == 'HP:0001742')]['measured'].values[0]
        self.assertTrue(measured)

    def test_excluded2(self):
        """
        The goal if this test is to enzure that 'Seizure' is annotated as 'excluded'
        """
        vignette = "The patient is not on seizure medication at this time."
        excluded = set()
        excluded.add("Seizure")
        encoder = CaseEncoder(concept_recognizer=self._hpo_cr, pmid="PMID:1")
        df = encoder.add_vignette(vignette=vignette, excluded_terms=excluded)
        self.assertEqual(1, len(df))
        
        self.assertEqual("HP:0001250", df.iloc[0]['id'])
        self.assertEqual("Seizure",df.iloc[0]['label'])
        self.assertFalse(df.iloc[0]['observed'])
        self.assertTrue(df.iloc[0]['measured'])





from typing import List
from .hp_term import HpTerm


class HpoConceptRecognizer:
    """
    _summary_
    This class acts as an interface for classes that implement parse_cell to perform HPO-based concept recognition.
    """
    def __init__(self):
        pass
    
    def parse_cell(self, cell_contents, custom_d=None)-> List[HpTerm]:
        raise NotImplementedError("Need to implement a subclass of HpoConceptRecognizer for CR")    

    def get_term_from_id(self, id) -> HpTerm:
        raise NotImplementedError("Need to implement a subclass of HpoConceptRecognizer for CR")   


import phenopackets


NA_STRING = "n/a"


class SimpleVariant:

    def __init__(self, interpretation) -> None:
        if str(type(interpretation)) != "<class 'phenopackets.schema.v2.core.interpretation_pb2.Interpretation'>":
            raise ValueError(f"interpretation argument must be GA4GH Phenopacket Interpretation but was {type(interpretation)}")
        diagnosis = interpretation.diagnosis
        # TODO CHECK FOR DISEASE
        ginterpretations = diagnosis.genomic_interpretations
        for gint in ginterpretations:
            self._status = str(gint.interpretation_status)
            vinterpretation = gint.variant_interpretation
            if vinterpretation.variation_descriptor is  None:
                raise ValueError("variation_descriptor was None")
            vdescript = vinterpretation.variation_descriptor
            if vdescript.gene_context is not None:
                self._gene_id = vdescript.gene_context.value_id
                self._gene_symbol = vdescript.gene_context.symbol
            else:
                self._gene_id = NA_STRING
                self._gene_symbol = NA_STRING
            if vdescript.expressions is not None and len(vdescript.expressions) > 0:
                for exprsn in vdescript.expressions:
                    syntax = exprsn.syntax
                    val = exprsn.value
                    if syntax == "hgvs.c":
                        self._hgvs = val
                    else:
                        self._hgvs = NA_STRING
            if vdescript.vcf_record is not None:
                vcf = vdescript.vcf_record
                self._genome_ass = vcf.genome_assembly
                self._chrom = vcf.chrom
                self._pos = int(vcf.pos)
                self._ref = vcf.ref
                self._alt = vcf.alt
            else:
                self._genome_ass = NA_STRING
                self._chrom = NA_STRING
                self._pos = -1
                self._ref = NA_STRING
                self._alt = NA_STRING
            if vdescript.allelic_state is not None:
                allelic = vdescript.allelic_state
                self._genotype_id = allelic.id
                self._genotype_label = allelic.label
            else:
                self._genotype_id = NA_STRING
                self._genotype_label = NA_STRING

    @property
    def status(self):
        return self._status

    @property
    def gene_id(self):
        return self._gene_id

    @property
    def gene_symbol(self):
        return self._gene_symbol
    
    @property
    def hgvs(self):
        return self._hgvs

    @property
    def genome_assembly(self):
        return self._genome_ass
    
    @property
    def chrom(self):
        return self._chrom

    @property
    def position(self):
        return self._pos

    @property
    def ref(self):
        return self._ref

    @property
    def alt(self):
        return self._alt

    @property
    def genotype_id(self):
        return self._genotype_id

    @property
    def genotype_label(self):
        return self._genotype_label

    def has_vcf(self):
        return self._chrom != NA_STRING

    def has_genotype(self):
        return self._genotype_id != NA_STRING


from datetime import datetime
from collections import defaultdict
import phenopackets 
from google import protobuf
import time





class Resource:
    def __init__(self, id, name, namespace_prefix, iriprefix, url, version) -> None:
        self._id = id
        self._name = name
        self._namespace_prefix = namespace_prefix
        self._iri_prefix = iriprefix
        self._url = url
        self._version = version
        
    def to_ga4gh(self):
        pass

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        return self._name

    @property
    def namespace_prefix(self):
        return self._namespace_prefix

    @property
    def iri_prefix(self):
        return self._iri_prefix

    @property
    def url(self):
        return self._url

    @property
    def version(self):
        return self._version


# current versions of ontologies used for these phenopackets
# if the user does not specify versions, these versions will be
# used.
# for our purposes, the one version that will change a lot is that
# of the HPO and users can get that from the HPO file and this should be
# provided in the constructor of metadata
        
default_versions = {
    'geno': '2022-03-05',
    'hgnc': '06/01/23',
    'omim': 'January 4, 2023',
    'mondo': 'v2023-01-04'
}  
   

class MetaData:
    """
    A representation of the MetaData element of the GA4GH Phenopacket Schema
    """
    
    def __init__(self, created_by) -> None:
        self._created_by = created_by
        self._schema_version = "2.0"
        self._resource_d = defaultdict(Resource)

    def default_versions_with_hpo(self, version):
        """
        Add resources for HPO (with specified version), GENO, HGNC, and OMIM (with default versions)
        The HPO version can be easily obtained from the HpoParser using the get_version() function
        """
        self.geno()
        self.hgnc()
        self.omim()
        self.hpo(version=version)
        
    def hpo(self, version):
        self._resource_d["hp"] = Resource(id="hp", 
                        name="human phenotype ontology",
                        namespace_prefix="HP",
                        iriprefix="http://purl.obolibrary.org/obo/HP_",
                        url="http://purl.obolibrary.org/obo/hp.owl",
                        version=version)

        
    def geno(self, version=default_versions.get('geno')):
        """_summary_
        GENO is used for three terms: homozygous, heterozygous, hemizygous
        For this reason, we use a default version, since we assume these terms will not change.
        Args:
            version (str, optional): GENO version. Defaults to "2022-03-05".

        Returns:
            _type_: Resource object representing GENO
        """
        self._resource_d["geno"] = Resource(id="geno", 
                        name="Genotype Ontology",
                        namespace_prefix="GENO",
                        iriprefix="http://purl.obolibrary.org/obo/GENO_",
                        url="http://purl.obolibrary.org/obo/geno.owl",
                        version=version)
        
    def hgnc(self, version=default_versions.get('hgnc')):
        self._resource_d["hgnc"] =  Resource(id="hgnc", 
                        name="HUGO Gene Nomenclature Committee",
                        namespace_prefix="HGNC",
                        iriprefix="https://www.genenames.org/data/gene-symbol-report/#!/hgnc_id/",
                        url="https://www.genenames.org",
                        version=version)
    
    
    def omim(self, version=default_versions.get('omim')):
        self._resource_d["omim"] = Resource(id="omim", 
                        name="An Online Catalog of Human Genes and Genetic Disorders",
                        namespace_prefix="OMIM",
                        iriprefix="https://www.omim.org/entry/",
                        url="https://www.omim.org",
                        version=version)
    
    def mondo(self, version=default_versions.get('mondo')):
        self._resource_d["mondo"] = Resource(id="mondo", 
                        name="Mondo Disease Ontology",
                        namespace_prefix="MONDO",
                        iriprefix="http://purl.obolibrary.org/obo/MONDO_",
                        url="http://purl.obolibrary.org/obo/mondo.obo",
                        version=version)

    def to_ga4gh(self):
        """
        Use a time stamp for the current instant
        """
        metadata = phenopackets.MetaData()
        metadata.created_by = self._created_by
        now = time.time()
        seconds = int(now)
        nanos = int((now - seconds) * 10**9)
        timestamp = protobuf.timestamp_pb2.Timestamp(seconds=seconds, nanos=nanos)
        metadata.created.CopyFrom(timestamp)
        metadata.phenopacket_schema_version = self._schema_version
        for _, resource in self._resource_d.items():
            res = phenopackets.Resource()
            res.id = resource.id
            res.name = resource.name
            res.namespace_prefix = resource.namespace_prefix
            res.iri_prefix = resource.iri_prefix
            res.url = resource.url
            res.version = resource.version
            metadata.resources.append(res)
        return metadata



          

    
    
 
        
        
        

 
    
    
    
  


from collections import defaultdict
from typing import List
import pandas as pd
from .column_mapper import ColumnMapper
from .hpo_cr import HpoConceptRecognizer


class CustomColumnMapper(ColumnMapper):
    def __init__(self, concept_recognizer, custom_map_d=None) -> None:
        if custom_map_d is None:
            self._custom_map_d = defaultdict()
        else:
            self._custom_map_d = custom_map_d
        if not isinstance(concept_recognizer, HpoConceptRecognizer):
            raise ValueError("concept_recognizer argument must be HpoConceptRecognizer but was {type(concept_recognizer)}")
        self._concept_recognizer = concept_recognizer


    def map_cell(self, cell_contents) -> List:
        """
        This method should not be used by client code. Instead, it is called
        by the CohortEncoder object, which should provide the id_to_primary_d, label_to_id_d
        objects representing the automatic HPO dictionaries. It is designed to take
        the contents of a single cell in a Supplemental Table and to parse out 
        HPO terms.
        For now, we use a naive implementation that searches for exact matches. Moving forward
        it should be possible to implement the algorithm of fenominal that removes stop words
        and searches for ontology concepts in which the remaining tokens occur in any order.
        """
        results = self._concept_recognizer.parse_cell(cell_contents=cell_contents, custom_d=self._custom_map_d)
        if results is None:
            return []
        else:
            return results


    def preview_column(self, column):
        if not isinstance(column, pd.Series):
            raise ValueError("column argument must be pandas Series, but was {type(column)}")
        preview_d = defaultdict(list)
        for index, value in column.items():
            preview_d[value]  = self.map_cell(str(value))
        dlist = []
        for k, v in preview_d.items():
            if v is None or len(v) == 0:
                terms = "n/a"
            else:
                terms = "; ".join([hpo.to_string() for hpo in v])
            dlist.append({"column": k, "terms": terms})
        return pd.DataFrame(dlist)
    
   
            
        


from .hp_term import HpTerm
from .column_mapper import ColumnMapper
from typing import List
import pandas as pd

class ThresholdedColumnMapper(ColumnMapper):
    def __init__(self, hpo_id, hpo_label, threshold, call_if_above, observed_code=None):
        self._hpo_id = hpo_id
        self._hpo_label = hpo_label
        if not isinstance(threshold, int) and not isinstance(threshold, float):
            raise ValueError(f"threshold argument must be integer or float but was {threshold}")
        if not isinstance(call_if_above, bool):
            raise ValueError(f"call_if_above argument must be True or False but was {call_if_above}")
        self._threshold = float(threshold) # transform ints to float for simplicity, it will not affect result
        self._call_if_above = call_if_above
        self._observed_code = observed_code
        
    def map_cell(self, cell_contents) -> List[HpTerm]:
        if isinstance(cell_contents, str):
            contents = cell_contents.strip()
            if contents == self._observed_code:
                return [HpTerm(id=self._hpo_id, label=self._hpo_label)] 
        elif isinstance(cell_contents, int):
            contents = cell_contents
        elif isinstance(cell_contents, float):
            contents = cell_contents
        else:
            raise ValueError(f"Malformed cell contents for ThresholdedColumnMapper: {cell_contents}, type={type(cell_contents)}")
        try:
            value = float(contents)
            if self._call_if_above:
                if value > self._threshold:
                    return [HpTerm(id=self._hpo_id, label=self._hpo_label)]
                else:
                    return [HpTerm(id=self._hpo_id, label=self._hpo_label, observed=False)]
            else:
                if value > self._threshold:
                    return [HpTerm(id=self._hpo_id, label=self._hpo_label, observed=False)]
                else:
                    return [HpTerm(id=self._hpo_id, label=self._hpo_label)]
        except Exception as exc:
             return [HpTerm(id=self._hpo_id, label=self._hpo_label, measured=False)]
        

    def preview_column(self, column):
        if not isinstance(column, pd.Series):
            raise ValueError("column argument must be pandas Series, but was {type(column)}")
        dlist = []
        for _, value in column.items():
            results  = self.map_cell(str(value))
            if len(results) > 0:
                hpterm = results[0]
                dlist.append({"term": hpterm.hpo_term_and_id, "status": hpterm.display_value})
        return pd.DataFrame(dlist)     
   


from .simple_patient import SimplePatient
import pandas as pd
from .hpo_category import HpoCategorySet
from hpotk.model import TermId
from hpotk.algorithm import get_ancestors
from collections import defaultdict
from hpotk.algorithm import exists_path

ALL_ROOT = TermId.from_curie("HP:0000001")
PHENOTYPIC_ABNORMALITY_ROOT = TermId.from_curie("HP:0000118")



class DetailedSupplTable:
    """
    This class intends to facilitate the creation of a detailed supplemental Excel file with the phenotypic findings for
    an entire cohort. Each attribute (id, sex, age, HPO terms) are shown on rows, and the individuals who make up the cohort
    are shown as columns.
    """

    def __init__(self, patient_d, hp_ontology) -> None:
        if not isinstance(patient_d, dict):
            raise ValueError(f"patient_d argument must be dictionary but was {type(patient_d)}")
        self._patient_d = patient_d
        self._hp_ontology = hp_ontology
        self._total_counts = defaultdict(int)
        sex_d = defaultdict(int)
        var_d = defaultdict(int)
        for pat_id, pat in self._patient_d.items():
            sex_d[pat.get_sex()] += 1
            hpo_terms = pat.get_observed_hpo_d()
            anc_set = set() # graph with ancestors induced by all terms of the patient
            for hp_id in hpo_terms.keys():
                # key is a string such as HP:0001234, value is an HpTerm object
                # we need to convert it to an object from hpo-toolkit because get_ancestors returns HpTerm objects
                hp_termid = TermId.from_curie(hp_id)
                ancs = get_ancestors(self._hp_ontology,hp_termid)
                anc_set.add(hp_termid)
                anc_set.update(ancs)
                self._total_counts[hp_id] += 1
            for hp_id in anc_set:
                if hp_id == ALL_ROOT or hp_id == PHENOTYPIC_ABNORMALITY_ROOT:
                    continue
                else:
                    self._total_counts_propagated[hp_id.value] += 1
            variant_d = pat.get_variant_d()
            for var in variant_d.keys():
                var_d[var] += 1

            # TODO figure out what to do with biallelic
        self._hpo_category_set = HpoCategorySet(ontology=hp_ontology)


    def get_table(self):
        """
        Get counts of terms within annotation propagation or thresholding
        """
        rows = []
        N = self._n_patients
        for hpid, total_count in self._total_counts.items():
            total_per = 100*total_count/N
            total_s = f"{total_count}/{N} ({total_per:.1f}%)"
            hpterm = self._ontology.get_term(hpid)
            cat = self.get_category(termid=hpid)
            focus_count = self._focus_counts.get(hpid, 0)
            other_count = self._non_focus_counts.get(hpid, 0)
            d = {'category': cat, 'term': hpterm.name, 'HP:id': hpid, 'focus' : focus_count, 'other': other_count, 'total': total_s, 'total_count': total_count}
            rows.append(d)
        df = pd.DataFrame(rows)
        df.set_index('category', inplace=True)
        return df.sort_values(['category', 'total_count'], ascending=[True, False])


import os
from collections import defaultdict
from .simple_patient import SimplePatient

class PhenopacketIngestor:
    
    def __init__(self, indir) -> None:
        if not os.path.isdir(indir):
            raise ValueError(f"indir argument {indir} must be directory!")
        self._indir = indir
        self._phenopackets = []
        self._patient_d = defaultdict(SimplePatient)
        for root, dirs, files in os.walk(indir):
            for file in files:
                if file.endswith(".json"):
                    self._phenopackets.append(file)
                    fullpath = os.path.join(indir, file)
                    patient = SimplePatient(phenopacket_file=fullpath)
                    phenopacket_id = patient.get_phenopacket_id()
                    self._patient_d[phenopacket_id] = patient

    def get_patient_dictionary(self):
        return self._patient_d
        


import re
import os
from typing import List
from collections import defaultdict
from .hp_term import HpTerm
from .hpo_cr import HpoConceptRecognizer
from google.protobuf.json_format import MessageToJson
from .individual import Individual
import phenopackets



ISO8601_REGEX = r"^P(\d+Y)?(\d+M)?(\d+D)?"





class CaseEncoder:
    
    
    def __init__(self, concept_recognizer, pmid, age_at_last_exam=None) -> None:
        if not isinstance(concept_recognizer, HpoConceptRecognizer):
            raise ValueError("concept_recognizer argument must be HpoConceptRecognizer but was {type(concept_recognizer)}")
        self._concept_recognizer = concept_recognizer
        if not pmid.startswith("PMID:"):
            raise ValueError(f"Malformed pmid argument ({pmid}). Must start with PMID:")
        self._pmid = pmid
        if age_at_last_exam is not None:
            match = re.search(ISO8601_REGEX, age_at_last_exam)
            if match:
                self._age_at_last_examination = age_at_last_exam
            else:
                raise ValueError(f"Could not parse {age_at_last_exam} as ISO8601 period")
        else:
            self._age_at_last_examination = None

        self._annotations = defaultdict(list)


    def add_vignette(self, vignette, custom_d=None, custom_age=None, false_positive = [], excluded_terms = set()) -> List[HpTerm]:
        """
        false_positive: words or phrases that should not be parsed to HP terms
        """
        if custom_d is None:
            custom_d = {}
        # replace new lines and multiple consecutive spaces with a single space
        text = re.sub('\s+',' ',vignette.replace('\n', ' '))
        for fp in false_positive:
            text = text.replace(fp, " ")
        # results will be a list with HpTerm elements
        results =  self._concept_recognizer._parse_chunk(chunk=text, custom_d=custom_d)
        if custom_age is not None:
            self._annotations[custom_age].extend(results)
        elif self._age_at_last_examination is not None:
            self._annotations[self._age_at_last_examination].extend(results)
        else:
            self._annotations["N/A"].extend(results)
        for r in results:
            if r.label in excluded_terms:
                r.excluded()
        return HpTerm.term_list_to_dataframe(results)
        
    def add_term(self, label=None, id=None, excluded=False, custom_age=None):
        if label is not None:
            results =  self._concept_recognizer._parse_chunk(chunk=label, custom_d={})
            if len(results) != 1:
                raise ValueError(f"Malformed label {label}  we got {len(results)} results")
            hpo_term = results[0]
            hpo_term._observed = not excluded
            if custom_age is not None:
                self._annotations[custom_age].append(hpo_term)
            elif self._age_at_last_examination is not None:
                self._annotations[self._age_at_last_examination].append(hpo_term)
            else:
                self._annotations["N/A"].append(hpo_term)
            return HpTerm.term_list_to_dataframe([hpo_term])
        elif id is not None:
            hpo_term = self._concept_recognizer.get_term_from_id(id)
            if excluded:
                hpo_term._observed = not excluded
            if custom_age is not None:
                self._annotations[custom_age].append(hpo_term)
            elif self._age_at_last_examination is not None:
                self._annotations[self._age_at_last_examination].append(hpo_term)
            else:
                self._annotations["N/A"].append(hpo_term)
            return HpTerm.term_list_to_dataframe([hpo_term])
        else:
            return ValueError("Must call function with non-None value for either id or label argument")


    def get_hpo_term_dict(self):
        return self._annotations

    def get_phenopacket(self, individual_id, metadata,  sex=None, age=None, disease_id=None, disease_label=None, variants=None):
        if not isinstance(variants, list):
            variants = [variants]
        individual = Individual(individual_id=individual_id, sex=sex, age=age, hpo_terms=self._annotations,
                       variant_list=variants, disease_id=disease_id, disease_label=disease_label)
        phenopacket_id = self._pmid.replace(":", "_") + "_" + individual.id.replace(" ", "_").replace(":", "_")
        return individual.to_ga4gh_phenopacket(metadata=metadata, phenopacket_id=phenopacket_id)
        

    def output_phenopacket(self, outdir, phenopacket):
        """_summary_

        Args:
            outdir (_type_): name of directory to write phenopackets
            individual_list (_type_, optional): List of individuals. If None, recreate
        """
        if not os.path.exists(outdir):
            os.makedirs(outdir)
        phenopacket_id = phenopacket.id
        json_string = MessageToJson(phenopacket)
        fname = phenopacket_id.replace(" ", "_") + ".json"
        outpth = os.path.join(outdir, fname)
        with open(outpth, "wt") as fh:
            fh.write(json_string)
            print(f"Wrote phenopacket to {outpth}")